/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::cmp::Ordering;
use std::collections::{HashMap, HashSet};

use futures::stream::FuturesUnordered;
use futures::StreamExt;
use matrix_sdk::async_trait;
use nonempty::NonEmpty;
use ratatui::text::Text;
use sorted_vec::SortedVec;

use crate::widgets::Prerender;

mod log;
pub use log::LogBuffer;
mod room;
pub use room::RoomBuffer;

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum BufferId {
  /// The main/home buffer
  Log,
  /// Any Matrix room
  Room(matrix_sdk::ruma::OwnedRoomId),
}

/// Values should follow the ["Ordering of children within a space" algorithm](https://spec.matrix.org/v1.8/client-server-api/#ordering-of-children-within-a-space).
#[derive(Clone, Debug, Default, Eq, PartialEq)]
pub struct BufferSortKey {
  /// Explicit `order` key in the `m.space.child` event
  explicit_order: Option<String>,
  origin_server_ts: Option<matrix_sdk::ruma::MilliSecondsSinceUnixEpoch>,
}

impl Ord for BufferSortKey {
  fn cmp(&self, other: &Self) -> Ordering {
    match (self.explicit_order.as_ref(), other.explicit_order.as_ref()) {
      (Some(e1), Some(e2)) => e1.cmp(&e2),
      (Some(_), None) => Ordering::Less, // children without 'order' go to the end
      (None, Some(_)) => Ordering::Greater, // ditto
      (None, None) => Ordering::Equal,
    }
    // Should we apply the same algo here? The algo from the Matrix spec does not apply
    // because it does not consider children with m.space.parent.
    // I'd say yes, so that all children declared by the parent are first, then all children
    // not explicitly recognized by the parent; instead of having the latter in the middle
    // of the list if only some children of the parent have an explicit "order" key.
    .then(
      match (
        self.origin_server_ts.as_ref(),
        other.origin_server_ts.as_ref(),
      ) {
        (Some(ts1), Some(ts2)) => ts1.cmp(&ts2),
        (Some(_), None) => Ordering::Less,
        (None, Some(_)) => Ordering::Greater,
        (None, None) => Ordering::Equal,
      },
    )
  }
}

impl PartialOrd for BufferSortKey {
  fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
    Some(self.cmp(other))
  }
}

#[derive(Debug, Clone)]
pub enum BufferItemContent<'buf> {
  Text(Text<'buf>),
  Divider(Text<'buf>),
  Empty,
}

pub struct BufferItem<'buf> {
  pub content: BufferItemContent<'buf>,
  pub prerender: &'buf Prerender,
}

#[async_trait]
pub trait Buffer: Send + Sync + memuse::DynamicUsage {
  /// A short human-readable name for the room, eg. to show in compact buflist
  fn short_name(&self) -> String;
  /// Unique identifier of this buffer
  fn id(&self) -> BufferId;
  /// Identifier of this buffer's parent buffer, if any.
  fn parent(&self) -> Option<BufferId> {
    None
  }
  fn children(&self) -> Option<SortedVec<(BufferSortKey, BufferId)>> {
    None
  }
  /// Returns if there are any updates to apply.
  async fn poll_updates(&mut self);
  fn content<'a>(&'a self) -> Box<dyn ExactSizeIterator<Item = BufferItem<'a>> + 'a>;
  /// Called when the user is being showned the oldest items this buffer returned.
  ///
  /// This should return immediately, not waiting for anything to be loaded.
  fn request_back_pagination(&self, num: u16) {}

  fn dynamic_usage(&self) -> usize {
    memuse::DynamicUsage::dynamic_usage(self)
  }
  fn dynamic_usage_bounds(&self) -> (usize, Option<usize>) {
    memuse::DynamicUsage::dynamic_usage_bounds(self)
  }
}

pub struct Buffers {
  buffers: Vec<Box<dyn Buffer>>,
  /// Set of children of each buffer, sorted so that children explicitly listed by a
  /// space are sorted according to
  /// https://spec.matrix.org/v1.8/client-server-api/#ordering-of-children-within-a-space
  /// and rooms that declare that space as their parent are sorted in a best-effort
  children: HashMap<BufferId, SortedVec<(BufferSortKey, BufferId)>>,

  /// Set of parents of each buffer.
  ///
  /// Ideally it should be unique, but it is possible for multiple spaces to list a room,
  /// and/or for rooms to declare multiple parents.
  /// In order to pick a parent deterministically, we use the smallest room id.
  parents: HashMap<BufferId, HashSet<BufferId>>,

  /// Set of buffers already placed after a parent space, so other spaces should not
  /// steal them, even if they are also their child (or it would cause buffers to move
  /// every time we re-sort the buffer list)
  attached_to_parent: HashSet<BufferId>,
  active_index: usize,
}

impl Buffers {
  pub fn new(initial_buffer: Box<dyn Buffer>) -> Self {
    Self {
      buffers: vec![initial_buffer],
      children: HashMap::new(),
      parents: HashMap::new(),
      attached_to_parent: HashSet::new(),
      active_index: 0,
    }
  }

  pub async fn poll_updates(&mut self) {
    self
      .iter_mut()
      .map(|buf| buf.poll_updates())
      .collect::<FuturesUnordered<_>>()
      .next()
      .await
      .expect("poll_updates reached the end of the never-ending stream");

    // Reorder buffers in case we just got an update on space relationships
    // FIXME: do this only when needed
    let mut buffers = Vec::new();
    //self.children.clear();
    //self.parents.clear();
    self.attached_to_parent.clear();
    std::mem::swap(&mut self.buffers, &mut buffers);
    for buf in buffers.into_iter() {
      self.push(buf);
    }
  }

  pub fn len(&self) -> usize {
    self.buffers.len()
  }

  pub fn parents(&self) -> &HashMap<BufferId, HashSet<BufferId>> {
    &self.parents
  }

  pub fn children(&self) -> &HashMap<BufferId, SortedVec<(BufferSortKey, BufferId)>> {
    &self.children
  }

  pub fn iter(&self) -> impl Iterator<Item = &dyn Buffer> {
    self.buffers.iter().map(|buffer_box| &**buffer_box)
  }

  pub fn iter_mut(&mut self) -> impl Iterator<Item = &mut Box<dyn Buffer>> {
    self.buffers.iter_mut()
  }

  /// `O(self.len())` if the buffer has a parent
  pub fn push(&mut self, new_buffer: Box<dyn Buffer>) {
    let new_buffer_id = new_buffer.id();

    match new_buffer.children() {
      Some(children) => {
        for child in children.iter() {
          self
            .parents
            .entry(child.1.clone())
            .or_insert_with(HashSet::new)
            .insert(new_buffer_id.clone());
        }
        let all_children = self
          .children
          .entry(new_buffer_id.clone())
          .or_insert_with(SortedVec::new);
        // all_children.extend() uses .insert() which always does a bisect search,
        // while .push is O(1) when the element is at the end (which is usually what
        // happens here).
        for child in children.into_vec().into_iter() {
          all_children.push(child);
        }
      },
      None => {},
    };
    let parent = match new_buffer.parent() {
      Some(parent) => {
        self
          .parents
          .entry(new_buffer_id.clone())
          .or_insert_with(HashSet::new)
          .insert(parent.clone());
        let siblings = self
          .children
          .entry(parent.clone())
          .or_insert_with(SortedVec::new);
        if siblings
          .iter()
          .find(|(_key, buffer_id)| *buffer_id == new_buffer_id)
          .is_none()
        {
          siblings.push((BufferSortKey::default(), new_buffer_id.clone()));
        }
        Some(parent)
      },
      None => self
        .parents
        .get(&new_buffer_id)
        .and_then(|parents| parents.iter().min().cloned()), // .min() for determinism
    };
    let children = self
      .children
      .entry(new_buffer_id.clone())
      .or_insert_with(SortedVec::new);

    match parent.and_then(|parent_id| self.buffers.iter().position(|buf| buf.id() == parent_id)) {
      None => self.buffers.push(new_buffer),
      Some(parent_position) => {
        // iterate through buffers after the parent, and record on the stack the chain
        // of successor from the parent to the current buffer. As soon as we see a buffer
        // with a parent not on the stack (or with no parent), it means we left the subtree,
        // and should insert the new buffer there.
        let mut stack = vec![self.buffers[parent_position].id()];
        let mut new_buffer = Some(new_buffer);
        for (i, buf) in self.buffers.iter().enumerate().skip(parent_position + 1) {
          match self
            .parents
            .entry(buf.id())
            .or_insert_with(HashSet::new)
            .iter()
            .flat_map(|parent_id| stack.iter().position(|id| *id == *parent_id))
            .min() // for determinism
          {
            None => {}, // root buffer, or parent is not in the subtree
            Some(parent_position) => {
              // parent is in the subtree
              stack.truncate(parent_position + 1);
              stack.push(buf.id());
              self.attached_to_parent.insert(new_buffer_id.clone());
              continue;
            },
          }
          // if we are here, it means buf.parent() is not in the subtree (or buf is orphan)
          // so we should insert the new buffer at this position
          self
            .buffers
            .insert(i, new_buffer.take().expect("new_buffer was already taken"));
          break;
        }

        if let Some(new_buffer) = new_buffer {
          // The parent's subtree was the last subtree
          self.buffers.push(new_buffer);
          self.attached_to_parent.insert(new_buffer_id.clone());
        }
      },
    }

    // Find existing children and move them after this buffer (while preserving their
    // relative order.

    // 1. pop children from the list of buffers
    // FIXME: should be recursive
    let mut all_buffers = Vec::new();
    let mut child_buffers = Vec::new();
    std::mem::swap(&mut all_buffers, &mut self.buffers);
    for buf in all_buffers.into_iter() {
      if children
        .iter()
        .find(|(_key, buf_id)| *buf_id == buf.id())
        .is_some()
        && !self.attached_to_parent.contains(&buf.id())
      {
        self.attached_to_parent.insert(buf.id());
        child_buffers.push(buf);
      } else {
        self.buffers.push(buf);
      }
    }

    // 2. split the set of buffers, and insert the children after the new buffer
    let after_children = self.buffers.split_off(
      self
        .buffers
        .iter()
        .position(|buf| buf.id() == new_buffer_id)
        .expect("new buffer disappeared")
        + 1,
    );
    self.buffers.extend(child_buffers.into_iter());
    self.buffers.extend(after_children);
  }

  pub fn active_index(&self) -> usize {
    self.active_index
  }

  pub fn set_active_index(&mut self, index: isize) {
    self.active_index = index.wrapping_rem_euclid(self.buffers.len() as isize) as usize;
  }

  pub fn active_buffer(&self) -> &dyn Buffer {
    &**self
      .buffers
      .get(self.active_index)
      .expect("Active buffer index does not exist")
  }

  pub fn active_buffer_mut(&mut self) -> &mut Box<dyn Buffer> {
    self
      .buffers
      .get_mut(self.active_index)
      .expect("Active buffer index does not exist")
  }
}
