/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use color_eyre::eyre::{Result, WrapErr};
use ratatui::{prelude::*, widgets::*};
use tokio::sync::mpsc::UnboundedSender;
use tokio::sync::OnceCell;

use super::Component;
use crate::{
  action::Action,
  config::{Config, KeyBindings},
};

#[derive(Default)]
pub struct Buflist {
  command_tx: Option<UnboundedSender<Action>>,
  config: OnceCell<Config>,
}
impl Buflist {
  pub fn new() -> Self {
    Self::default()
  }
}

impl Component for Buflist {
  fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
    self.command_tx = Some(tx);
    Ok(())
  }

  fn register_config_handler(&mut self, config: Config) -> Result<()> {
    self
      .config
      .set(config)
      .context("Buflist config was already set")
  }

  fn update(&mut self, action: &Action) -> Result<Option<Action>> {
    match action {
      Action::Tick => {},
      _ => {},
    }
    Ok(None)
  }

  fn draw(
    &mut self,
    frame: &mut Frame<'_>,
    area: Rect,
    buffers: &crate::buffers::Buffers,
  ) -> Result<()> {
    let num_digits = u32::min(5, buffers.len().ilog10() + 1) as usize;
    let right_pad = " ".repeat(area.width.into());
    let mut stack = Vec::new(); // List of parent buffers of the current one
    frame.render_widget(
      Paragraph::new(
        buffers
          .iter()
          .enumerate()
          .map(|(i, buf)| {
            match buffers
              .parents()
              .get(&buf.id())
              .into_iter()
              .flatten()
              .flat_map(|parent| stack.iter().position(|id| *id == *parent))
              .max() // if in multiple spaces, the last one is most likely to matter
            {
              Some(parent_position) => stack.truncate(parent_position + 1),
              None => stack.clear(),
            }
            stack.push(buf.id());
            let buf_number = format!("{}.", i + 1);
            let mut base_style = Style::default();
            if i == buffers.active_index() {
              base_style = base_style.on_blue();
            }
            let tree_pad = " ".repeat(stack.len() - 1);
            let buf_number_style = base_style.green();
            let left_pad = " ".repeat((num_digits + 1).saturating_sub(buf_number.len()));
            vec![
              Span::styled(left_pad, base_style),
              Span::styled(buf_number, buf_number_style),
              Span::styled(tree_pad, base_style),
              Span::styled(buf.short_name(), base_style),
              Span::styled(right_pad.clone(), base_style),
            ]
            .into()
          })
          .collect::<Vec<Line<'_>>>(),
      )
      .block(Block::new().borders(Borders::ALL)),
      area,
    );

    Ok(())
  }
}
