/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::{collections::HashMap, time::Duration};

use color_eyre::eyre::{Result, WrapErr};
use crossterm::event::{KeyCode, KeyEvent, KeyEventKind, KeyModifiers};
use ratatui::{prelude::*, widgets::*};
use serde::{Deserialize, Serialize};
use tokio::sync::mpsc::UnboundedSender;
use tokio::sync::OnceCell;
use tui_textarea::{Input, Key, TextArea};

use super::{Backlog, Buflist, Component};
use crate::{
  action::Action,
  config::{Config, KeyBindings},
};

#[derive(Default)]
pub struct Home {
  command_tx: Option<UnboundedSender<Action>>,
  config: OnceCell<Config>,
  buflist: Buflist,
  backlog: Backlog,
  textarea: TextArea<'static>,
}

impl Home {
  pub fn new() -> Self {
    let mut self_ = Self::default();
    self_.configure_textarea();
    self_
  }

  fn configure_textarea(&mut self) {
    self
      .textarea
      .set_block(Block::default().borders(Borders::ALL));
    self.textarea.set_cursor_line_style(Style::default());
  }
}

impl Component for Home {
  fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
    self.buflist.register_action_handler(tx.clone())?;
    self.command_tx = Some(tx);
    Ok(())
  }

  fn register_config_handler(&mut self, config: Config) -> Result<()> {
    self.buflist.register_config_handler(config.clone())?;
    self
      .config
      .set(config)
      .context("Home config was already set")?;
    Ok(())
  }

  fn handle_key_events(&mut self, key: KeyEvent) -> Result<Option<Action>> {
    match key {
      KeyEvent {
        code: KeyCode::Enter,
        modifiers: KeyModifiers::NONE,
        kind: KeyEventKind::Press,
        state: _,
      } => {
        // User pressed enter; clear the textarea by replacing it with a new one,
        // and get the content of the old one.
        let mut textarea = TextArea::default();
        std::mem::swap(&mut textarea, &mut self.textarea);
        let command = textarea.into_lines().join("\n");
        self.configure_textarea();
        Ok(Some(Action::RunCommand(command)))
      },
      KeyEvent {
        code: KeyCode::Enter,
        modifiers: KeyModifiers::SHIFT,
        kind: KeyEventKind::Press,
        state: _,
      } => {
        // FIXME: For some reason this does nothing because Crossterm does not emit
        // Shift-Enter events (at least on Foot via SSH).
        // However, tui-textarea implements Alt-Tab to insert a newline, so it's okay.
        self.textarea.insert_newline();
        Ok(Some(Action::ShouldRender))
      },
      _ => {
        assert!(
          self.backlog.handle_key_events(key)?.is_none(),
          "backlog.handle_key_events returned Some"
        );
        self.textarea.input(key);
        Ok(Some(Action::ShouldRender))
      },
    }
  }

  fn update(&mut self, action: &Action) -> Result<Option<Action>> {
    self.buflist.update(action)?;
    match action {
      Action::Tick => {},
      _ => {},
    }
    Ok(None)
  }

  fn draw(
    &mut self,
    frame: &mut Frame<'_>,
    area: Rect,
    buffers: &crate::buffers::Buffers,
  ) -> Result<()> {
    let layout = Layout::default()
      .direction(Direction::Horizontal)
      .constraints(vec![Constraint::Percentage(20), Constraint::Percentage(80)])
      .split(area);

    self
      .buflist
      .draw(frame, layout[0], buffers)
      .context("Error drawing buflist")?;

    let layout = Layout::default()
      .direction(Direction::Vertical)
      .constraints(vec![
        Constraint::Min(5),
        Constraint::Length(self.textarea.lines().len() as u16 + 2), // +2 for borders
      ])
      .split(layout[1]);

    frame.render_widget(self.textarea.widget(), layout[1]);

    self
      .backlog
      .draw(frame, layout[0], buffers)
      .context("Error drawing backlog")?;

    Ok(())
  }
}
