/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use color_eyre::eyre::{Result, WrapErr};
use itertools::Itertools;
use tokio::sync::mpsc;

use crate::action::Action;
use crate::commands::{CommandHandler, RataCommand};
use crate::plugins::{Plugin, PluginGetter, PrePlugin};

inventory::submit! {
  PluginGetter(get_plugin)
}

fn get_plugin() -> Result<Box<dyn PrePlugin>> {
  Ok(Box::new(Memuse {}))
}

struct Memuse {}

impl PrePlugin for Memuse {
  fn name(&self) -> String {
    "Memuse".to_owned()
  }

  fn help(&self) -> String {
    "estimates memory usage".to_owned()
  }

  fn load(self: Box<Self>, app: &mut crate::App) -> Result<Box<dyn Plugin>> {
    app.commands.register(Box::new(MemuseCommand {}));
    Ok(self)
  }
}

impl Plugin for Memuse {}

struct MemuseCommand {}

impl RataCommand for MemuseCommand {
  fn name(&self) -> String {
    "memuse".to_owned()
  }
  fn help(&self) -> String {
    "returns memory usage estimates".to_owned()
  }
  fn handler(&self) -> CommandHandler {
    Box::new(move |_name, _args, app, action_tx| {
      /*
      let client_usage = app
        .clients
        .iter()
        .map(|client| format_mem_usage(client.user_id().map(|user_id| user_id.as_str()).unwrap_or("<INITIALIZING>"), client))
        .join("\n");*/
      let client_usage = "todo";
      let buffer_usage = app
        .buffers
        .iter()
        .map(|buffer| format_mem_usage2(&buffer.short_name(), buffer))
        .join("\n");
      tracing::info!(
        "Client memory usage (matrix-sdk):\n{}\nBuffer memory usage:\n{}",
        client_usage,
        buffer_usage
      );
      Ok(())
    })
  }
}

fn format_mem_usage<T: memuse::DynamicUsage>(name: &str, value: T) -> String {
  match value.dynamic_usage_bounds() {
    (min, Some(max)) => format!(
      "* {}: {} to {} bytes, approximately {}",
      name,
      min,
      max,
      value.dynamic_usage()
    ),
    (min, None) => format!(
      "* {}: approximately {} bytes but no less than {}",
      name,
      min,
      value.dynamic_usage()
    ),
  }
}

/// Boo trait objects
fn format_mem_usage2(name: &str, value: &dyn crate::buffers::Buffer) -> String {
  match crate::buffers::Buffer::dynamic_usage_bounds(value) {
    (min, Some(max)) => format!(
      "* {}: {} to {} bytes, approximately {}",
      name,
      min,
      max,
      crate::buffers::Buffer::dynamic_usage(value)
    ),
    (min, None) => format!(
      "* {}: approximately {} bytes but no less than {}",
      name,
      min,
      crate::buffers::Buffer::dynamic_usage(value)
    ),
  }
}
